@@ -18,7 +18,7 @@
  */
 
 /***************************************************************************
- * Copyright (C) 2013-2015 Ping Identity Corporation
+ * Copyright (C) 2013-2016 Ping Identity Corporation
  * All rights reserved.
  *
  * For further information please contact:
@@ -324,7 +324,7 @@ static apr_byte_t oidc_unsolicited_proto_state(request_rec *r, oidc_cfg *c,
 		return FALSE;
 	}
 
-	if (strcmp(rfp, "iss") != 0) {
+	if (apr_strnatcmp(rfp, "iss") != 0) {
 		oidc_error(r, "\"rfp\" (%s) does not match \"iss\", aborting", rfp);
 		apr_jwt_destroy(jwt);
 		return FALSE;
@@ -412,6 +412,59 @@ static apr_byte_t oidc_unsolicited_proto_state(request_rec *r, oidc_cfg *c,
 	return TRUE;
 }
 
+static json_t * oidc_get_state_from_cookie(request_rec *r, const char *cookieValue) {
+	json_t *result = NULL;
+
+	/* decrypt the state obtained from the cookie */
+	char *svalue = NULL;
+	if (oidc_base64url_decode_decrypt_string(r, &svalue, cookieValue) <= 0)
+		return NULL;
+
+	oidc_debug(r, "restored JSON state cookie value: %s", svalue);
+
+	json_error_t json_error;
+	result = json_loads(svalue, 0, &json_error);
+	if (result == NULL) {
+		oidc_error(r, "parsing JSON (json_loads) failed: %s", json_error.text);
+		return NULL;
+	}
+
+	return result;
+}
+
+static void oidc_clean_expired_state_cookies(request_rec *r, oidc_cfg *c) {
+	char *cookie, *tokenizerCtx;
+	char *cookies = apr_pstrdup(r->pool,
+			(char *) apr_table_get(r->headers_in, "Cookie"));
+	if (cookies != NULL) {
+		cookie = apr_strtok(cookies, ";", &tokenizerCtx);
+		do {
+			while (cookie != NULL && *cookie == ' ')
+				cookie++;
+			if (strstr(cookie, OIDCStateCookiePrefix) == cookie) {
+				char *cookieName = cookie;
+				while (cookie != NULL && *cookie != '=')
+					cookie++;
+				if (*cookie == '=') {
+					*cookie = '\0';
+					cookie++;
+					json_t *state = oidc_get_state_from_cookie(r, cookie);
+					if (state != NULL) {
+						json_t *v = json_object_get(state, "timestamp");
+						apr_time_t now = apr_time_sec(apr_time_now());
+						if (now > json_integer_value(v) + c->state_timeout) {
+							oidc_error(r, "state has expired");
+							oidc_util_set_cookie(r, cookieName, "", 0);
+						}
+						json_decref(state);
+					}
+				}
+			}
+			cookie = apr_strtok(NULL, ";", &tokenizerCtx);
+		} while (cookie != NULL);
+	}
+}
+
 /*
  * restore the state that was maintained between authorization request and response in an encrypted cookie
  */
@@ -420,6 +473,9 @@ static apr_byte_t oidc_restore_proto_state(request_rec *r, oidc_cfg *c,
 
 	oidc_debug(r, "enter");
 
+	/* clean expired state cookies to avoid pollution */
+	oidc_clean_expired_state_cookies(r, c);
+
 	const char *cookieName = oidc_get_state_cookie_name(r, state);
 
 	/* get the state cookie value first */
@@ -432,19 +488,8 @@ static apr_byte_t oidc_restore_proto_state(request_rec *r, oidc_cfg *c,
 	/* clear state cookie because we don't need it anymore */
 	oidc_util_set_cookie(r, cookieName, "", 0);
 
-	/* decrypt the state obtained from the cookie */
-	char *svalue = NULL;
-	if (oidc_base64url_decode_decrypt_string(r, &svalue, cookieValue) <= 0)
-		return FALSE;
-
-	oidc_debug(r, "restored JSON state cookie value: %s", svalue);
-
-	json_error_t json_error;
-	*proto_state = json_loads(svalue, 0, &json_error);
-	if (*proto_state == NULL) {
-		oidc_error(r, "parsing JSON (json_loads) failed: %s", json_error.text);
-		return FALSE;
-	}
+	*proto_state = oidc_get_state_from_cookie(r, cookieValue);
+	if (*proto_state == NULL) return FALSE;
 
 	json_t *v = json_object_get(*proto_state, "nonce");
 
@@ -469,6 +514,9 @@ static apr_byte_t oidc_restore_proto_state(request_rec *r, oidc_cfg *c,
 		return FALSE;
 	}
 
+	/* add the state */
+	json_object_set_new(*proto_state, "state", json_string(state));
+
 	char *s_value = json_dumps(*proto_state, JSON_ENCODE_ANY);
 	oidc_debug(r, "restored state: %s", s_value);
 	free(s_value);
@@ -498,12 +546,14 @@ static apr_byte_t oidc_authorization_request_set_cookie(request_rec *r,
 		return FALSE;
 	}
 
+	/* clean expired state cookies to avoid pollution */
+	oidc_clean_expired_state_cookies(r, c);
+
 	/* assemble the cookie name for the state cookie */
 	const char *cookieName = oidc_get_state_cookie_name(r, state);
 
 	/* set it as a cookie */
-	oidc_util_set_cookie(r, cookieName, cookieValue,
-			apr_time_now() + apr_time_from_sec(c->state_timeout));
+	oidc_util_set_cookie(r, cookieName, cookieValue, -1);
 
 	free(s_value);
 
@@ -643,8 +693,8 @@ static int oidc_check_max_session_duration(request_rec *r, oidc_cfg *cfg,
 		oidc_warn(r, "maximum session duration exceeded for user: %s",
 				session->remote_user);
 		oidc_session_kill(r, session);
-		return oidc_authenticate_user(r, cfg, NULL,
-				oidc_get_current_url(r, cfg), NULL,
+		return oidc_authenticate_user(r, cfg, NULL, oidc_get_current_url(r),
+				NULL,
 				NULL, NULL, NULL);
 	}
 
@@ -654,10 +704,265 @@ static int oidc_check_max_session_duration(request_rec *r, oidc_cfg *cfg,
 	return OK;
 }
 
+/*
+ * validate received session cookie against the domain it was issued for:
+ *
+ * this handles the case where the cache configured is a the same single memcache, Redis, or file
+ * backend for different (virtual) hosts, or a client-side cookie protected with the same secret
+ *
+ * it also handles the case that a cookie is unexpectedly shared across multiple hosts in
+ * name-based virtual hosting even though the OP(s) would be the same
+ */
+static apr_byte_t oidc_check_cookie_domain(request_rec *r, oidc_cfg *cfg,
+		session_rec *session) {
+	const char *c_cookie_domain =
+			cfg->cookie_domain ?
+					cfg->cookie_domain : oidc_get_current_url_host(r);
+	const char *s_cookie_domain = NULL;
+	oidc_session_get(r, session, OIDC_COOKIE_DOMAIN_SESSION_KEY,
+			&s_cookie_domain);
+	if ((s_cookie_domain == NULL)
+			|| (apr_strnatcmp(c_cookie_domain, s_cookie_domain) != 0)) {
+		oidc_warn(r,
+				"aborting: detected attempt to play cookie against a different domain/host than issued for! (issued=%s, current=%s)",
+				s_cookie_domain, c_cookie_domain);
+		return FALSE;
+	}
+
+	return TRUE;
+}
+
+/*
+ * get a handle to the provider configuration via the "issuer" stored in the session
+ */
+apr_byte_t oidc_get_provider_from_session(request_rec *r, oidc_cfg *c, session_rec *session, oidc_provider_t **provider) {
+
+	oidc_debug(r, "enter");
+
+	/* get the issuer value from the session state */
+	const char *issuer = NULL;
+	oidc_session_get(r, session, OIDC_ISSUER_SESSION_KEY, &issuer);
+	if (issuer == NULL) {
+		oidc_error(r, "session corrupted: no issuer found in session");
+		return FALSE;
+	}
+
+	/* get the provider info associated with the issuer value */
+	oidc_provider_t *p = oidc_get_provider_for_issuer(r, c, issuer, FALSE);
+	if (p == NULL) {
+		oidc_error(r, "session corrupted: no provider found for issuer: %s",
+				issuer);
+		return FALSE;
+	}
+
+	*provider = p;
+
+	return TRUE;
+}
+
+/*
+ * store the access token expiry timestamp in the session, based on the expires_in
+ */
+static void oidc_store_access_token_expiry(request_rec *r, session_rec *session,
+		int expires_in) {
+	if (expires_in != -1) {
+		oidc_session_set(r, session, OIDC_ACCESSTOKEN_EXPIRES_SESSION_KEY,
+				apr_psprintf(r->pool, "%" APR_TIME_T_FMT,
+						apr_time_sec(apr_time_now()) + expires_in));
+	}
+}
+
+/*
+ * store claims resolved from the userinfo endpoint in the session
+ */
+static void oidc_store_userinfo_claims(request_rec *r, session_rec *session,
+		oidc_provider_t *provider, const char *claims) {
+	/* see if we've resolved any claims */
+	if (claims != NULL) {
+		/*
+		 * Successfully decoded a set claims from the response so we can store them
+		 * (well actually the stringified representation in the response)
+		 * in the session context safely now
+		 */
+		oidc_session_set(r, session, OIDC_CLAIMS_SESSION_KEY, claims);
+
+		/* store the last refresh time if we've configured a userinfo refresh interval */
+		if (provider->userinfo_refresh_interval > 0)
+			oidc_session_set(r, session, OIDC_USERINFO_LAST_REFRESH_SESSION_KEY,
+					apr_psprintf(r->pool, "%" APR_TIME_T_FMT, apr_time_now()));
+	}
+}
+
+/*
+ * execute refresh token grant to refresh the existing access token
+ */
+static apr_byte_t oidc_refresh_access_token(request_rec *r, oidc_cfg *c,
+		session_rec *session, oidc_provider_t *provider,
+		char **new_access_token) {
+
+	oidc_debug(r, "enter");
+
+	/* get the refresh token that was stored in the session */
+	const char *refresh_token = NULL;
+	oidc_session_get(r, session, OIDC_REFRESHTOKEN_SESSION_KEY, &refresh_token);
+	if (refresh_token == NULL) {
+		oidc_warn(r,
+				"refresh token routine called but no refresh_token found in the session");
+		return FALSE;
+	}
+
+	/* elements returned in the refresh response */
+	char *s_id_token = NULL;
+	int expires_in = -1;
+	char *s_token_type = NULL;
+	char *s_access_token = NULL;
+	char *s_refresh_token = NULL;
+
+	/* refresh the tokens by calling the token endpoint */
+	if (oidc_proto_refresh_request(r, c, provider, refresh_token, &s_id_token,
+			&s_access_token, &s_token_type, &expires_in,
+			&s_refresh_token) == FALSE) {
+		oidc_error(r, "access_token could not be refreshed");
+		return FALSE;
+	}
+
+	/* store the new access_token in the session and discard the old one */
+	oidc_session_set(r, session, OIDC_ACCESSTOKEN_SESSION_KEY, s_access_token);
+	oidc_store_access_token_expiry(r, session, expires_in);
+
+	/* see if we need to return it as a parameter */
+	if (new_access_token != NULL)
+		*new_access_token = s_access_token;
+
+	/* if we have a new refresh token (rolling refresh), store it in the session and overwrite the old one */
+	if (s_refresh_token != NULL)
+		oidc_session_set(r, session, OIDC_REFRESHTOKEN_SESSION_KEY,
+				s_refresh_token);
+
+	return TRUE;
+}
+
+/*
+ * retrieve claims from the userinfo endpoint and return the stringified response
+ */
+static const char *oidc_retrieve_claims_from_userinfo_endpoint(request_rec *r,
+		oidc_cfg *c, oidc_provider_t *provider, const char *access_token, session_rec *session) {
+
+	oidc_debug(r, "enter");
+
+	/* see if a userinfo endpoint is set, otherwise there's nothing to do for us */
+	if (provider->userinfo_endpoint_url == NULL) {
+		oidc_debug(r,
+				"not retrieving userinfo claims because userinfo_endpoint is not set");
+		return NULL;
+	}
+
+	/* see if there's an access token, otherwise we can't call the userinfo endpoint at all */
+	if (access_token == NULL) {
+		oidc_debug(r,
+				"not retrieving userinfo claims because access_token is not provided");
+		return NULL;
+	}
+
+	// TODO: return code should indicate whether the token expired or some other error occurred
+	// TODO: long-term: session storage should be JSON (with explicit types and less conversion, using standard routines)
+
+	/* try to get claims from the userinfo endpoint using the provided access token */
+	const char *result = NULL;
+	if (oidc_proto_resolve_userinfo(r, c, provider, access_token, &result) == FALSE) {
+
+		/* see if we have an existing session and we are refreshing the user info claims */
+		if (session != NULL) {
+
+			/* first call to user info endpoint failed, but the access token may have just expired, so refresh it */
+			char *access_token = NULL;
+			if (oidc_refresh_access_token(r, c, session, provider, &access_token) == TRUE) {
+
+				/* try again with the new access token */
+				if (oidc_proto_resolve_userinfo(r, c, provider, access_token, &result) == FALSE) {
+
+					oidc_error(r,
+						"resolving user info claims with the refreshed access token failed, nothing will be stored in the session");
+					result = NULL;
+
+				}
+
+			} else {
+
+				oidc_warn(r,
+					"refreshing access token failed, claims will not be retrieved/refreshed from the userinfo endpoint");
+				result = NULL;
+
+			}
+
+		} else {
+
+			oidc_error(r,
+				"resolving user info claims with the existing/provided access token failed, nothing will be stored in the session");
+			result = NULL;
+
+		}
+	}
+
+	return result;
+}
+
+/*
+ * get (new) claims from the userinfo endpoint
+ */
+static apr_byte_t oidc_refresh_claims_from_userinfo_endpoint(request_rec *r,
+		oidc_cfg *cfg, session_rec *session) {
+
+	oidc_provider_t *provider = NULL;
+	const char *claims = NULL;
+	char *access_token = NULL;
+
+	/* get the current provider info */
+	if (oidc_get_provider_from_session(r, cfg, session, &provider) == FALSE)
+		return FALSE;
+
+	/* see if we can do anything here, i.e. we have a userinfo endpoint and a refresh interval is configured */
+	apr_time_t interval = apr_time_from_sec(
+			provider->userinfo_refresh_interval);
+
+	oidc_debug(r, "userinfo_endpoint=%s, interval=%d", provider->userinfo_endpoint_url, provider->userinfo_refresh_interval);
+
+	if ((provider->userinfo_endpoint_url != NULL) && (interval > 0)) {
+
+		/* get the last refresh timestamp from the session info */
+		apr_time_t last_refresh;
+		const char *s_last_refresh = NULL;
+		oidc_session_get(r, session, OIDC_USERINFO_LAST_REFRESH_SESSION_KEY,
+				&s_last_refresh);
+		sscanf(s_last_refresh, "%" APR_TIME_T_FMT, &last_refresh);
+
+		oidc_debug(r, "refresh needed in: %" APR_TIME_T_FMT " seconds", apr_time_sec(last_refresh + interval - apr_time_now()));
+
+		/* see if we need to refresh again */
+		if (last_refresh + interval < apr_time_now()) {
+
+			/* get the current access token */
+			oidc_session_get(r, session, OIDC_ACCESSTOKEN_SESSION_KEY,
+					(const char **) &access_token);
+
+			/* retrieve the current claims */
+			claims = oidc_retrieve_claims_from_userinfo_endpoint(r, cfg,
+					provider, access_token, session);
+
+			/* store claims resolved from userinfo endpoint */
+			oidc_store_userinfo_claims(r, session, provider, claims);
+
+			/* indicated something changed */
+			return TRUE;
+		}
+	}
+	return FALSE;
+}
+
 /*
  * handle the case where we have identified an existing authentication session for a user
  */
-static int oidc_handle_existing_session(request_rec *r, oidc_cfg * cfg,
+static int oidc_handle_existing_session(request_rec *r, oidc_cfg *cfg,
 		session_rec *session) {
 
 	oidc_debug(r, "enter");
@@ -666,11 +971,19 @@ static int oidc_handle_existing_session(request_rec *r, oidc_cfg * cfg,
 	oidc_dir_cfg *dir_cfg = ap_get_module_config(r->per_dir_config,
 			&auth_openidc_module);
 
+	/* verify current cookie domain against issued cookie domain */
+	if (oidc_check_cookie_domain(r, cfg, session) == FALSE)
+		return HTTP_UNAUTHORIZED;
+
 	/* check if the maximum session duration was exceeded */
 	int rc = oidc_check_max_session_duration(r, cfg, session);
 	if (rc != OK)
 		return rc;
 
+	/* if needed, refresh claims from the user info endpoint */
+	apr_byte_t needs_save = oidc_refresh_claims_from_userinfo_endpoint(r, cfg,
+			session);
+
 	/*
 	 * we're going to pass the information that we have to the application,
 	 * but first we need to scrub the headers that we're going to use for security reasons
@@ -692,11 +1005,8 @@ static int oidc_handle_existing_session(request_rec *r, oidc_cfg * cfg,
 	}
 
 	/* set the user authentication HTTP header if set and required */
-	if ((r->user != NULL) && (dir_cfg->authn_header != NULL)) {
-		oidc_debug(r, "setting authn header (%s) to: %s", dir_cfg->authn_header,
-				r->user);
-		apr_table_set(r->headers_in, dir_cfg->authn_header, r->user);
-	}
+	if ((r->user != NULL) && (dir_cfg->authn_header != NULL))
+		oidc_util_set_header(r, dir_cfg->authn_header, r->user);
 
 	/* set the claims in the app headers + request state */
 	if (oidc_set_app_claims(r, cfg, session, OIDC_CLAIMS_SESSION_KEY) == FALSE)
@@ -771,9 +1081,13 @@ static int oidc_handle_existing_session(request_rec *r, oidc_cfg * cfg,
 		slack = apr_time_from_sec(60);
 	if (session->expiry - now < interval - slack) {
 		session->expiry = now + interval;
-		oidc_session_save(r, session);
+		needs_save = TRUE;
 	}
 
+	/* check if something was updated in the session and we need to save it again */
+	if (needs_save)
+		oidc_session_save(r, session);
+
 	/* return "user authenticated" status */
 	return OK;
 }
@@ -867,22 +1181,9 @@ static int oidc_authorization_response_error(request_rec *r, oidc_cfg *c,
 	if ((prompt != NULL) && (apr_strnatcmp(prompt, "none") == 0)) {
 		return oidc_session_redirect_parent_window_to_logout(r, c);
 	}
-	return oidc_util_html_send_error(r,
-			apr_psprintf(r->pool,
-					"The OpenID Connect Provider returned an error: %s", error),
-					error_description, DONE);
-}
-
-/*
- * store the access token expiry timestamp in the session, based on the expires_in
- */
-static void oidc_store_access_token_expiry(request_rec *r, session_rec *session,
-		int expires_in) {
-	if (expires_in != -1) {
-		oidc_session_set(r, session, OIDC_ACCESSTOKEN_EXPIRES_SESSION_KEY,
-				apr_psprintf(r->pool, "%" APR_TIME_T_FMT,
-						apr_time_sec(apr_time_now()) + expires_in));
-	}
+	return oidc_util_html_send_error(r, c->error_template,
+			apr_psprintf(r->pool, "OpenID Connect Provider error: %s", error),
+			error_description, DONE);
 }
 
 /*
@@ -974,21 +1275,17 @@ static void oidc_save_in_session(request_rec *r, oidc_cfg *c,
 				provider->check_session_iframe);
 		oidc_session_set(r, session, OIDC_CLIENTID_SESSION_KEY,
 				provider->client_id);
+		oidc_debug(r, "session management enabled: stored session_state (%s), check_session_iframe (%s) and client_id (%s) in the session", session_state, provider->check_session_iframe, provider->client_id);
+	} else {
+		oidc_debug(r, "session management disabled: session_state (%s) and/or check_session_iframe (%s) is not provided", session_state, provider->check_session_iframe);
 	}
-
+	
 	if (provider->end_session_endpoint != NULL)
 		oidc_session_set(r, session, OIDC_LOGOUT_ENDPOINT_SESSION_KEY,
 				provider->end_session_endpoint);
 
-	/* see if we've resolved any claims */
-	if (claims != NULL) {
-		/*
-		 * Successfully decoded a set claims from the response so we can store them
-		 * (well actually the stringified representation in the response)
-		 * in the session context safely now
-		 */
-		oidc_session_set(r, session, OIDC_CLAIMS_SESSION_KEY, claims);
-	}
+	/* store claims resolved from userinfo endpoint */
+	oidc_store_userinfo_claims(r, session, provider, claims);
 
 	/* see if we have an access_token */
 	if (access_token != NULL) {
@@ -1018,6 +1315,10 @@ static void oidc_save_in_session(request_rec *r, oidc_cfg *c,
 	/* log message about max session duration */
 	oidc_log_session_expires(r, session_expires);
 
+	/* store the domain for which this session is valid */
+	oidc_session_set(r, session, OIDC_COOKIE_DOMAIN_SESSION_KEY,
+			c->cookie_domain ? c->cookie_domain : oidc_get_current_url_host(r));
+
 	/* store the session */
 	oidc_session_save(r, session);
 }
@@ -1090,27 +1391,6 @@ static apr_byte_t oidc_handle_flows(request_rec *r, oidc_cfg *c,
 	return rc;
 }
 
-/*
- * resolves claims from the user info endpoint and returns the stringified response
- */
-static const char *oidc_resolve_claims_from_user_info_endpoint(request_rec *r,
-		oidc_cfg *c, oidc_provider_t *provider, apr_table_t *params) {
-	const char *result = NULL;
-	if (provider->userinfo_endpoint_url == NULL) {
-		oidc_debug(r,
-				"not resolving user info claims because userinfo_endpoint is not set");
-	} else if (apr_table_get(params, "access_token") == NULL) {
-		oidc_debug(r,
-				"not resolving user info claims because access_token is not provided");
-	} else if (oidc_proto_resolve_userinfo(r, c, provider,
-			apr_table_get(params, "access_token"), &result) == FALSE) {
-		oidc_debug(r,
-				"resolving user info claims failed, nothing will be stored in the session");
-		result = NULL;
-	}
-	return result;
-}
-
 /* handle the browser back on an authorization response */
 static apr_byte_t oidc_handle_browser_back(request_rec *r, const char *r_state,
 		session_rec *session) {
@@ -1199,8 +1479,8 @@ static int oidc_handle_authorization_response(request_rec *r, oidc_cfg *c,
 	 * optionally resolve additional claims against the userinfo endpoint
 	 * parsed claims are not actually used here but need to be parsed anyway for error checking purposes
 	 */
-	const char *claims = oidc_resolve_claims_from_user_info_endpoint(r, c,
-			provider, params);
+	const char *claims = oidc_retrieve_claims_from_userinfo_endpoint(r,
+			c, provider, apr_table_get(params, "access_token"), NULL);
 
 	/* restore the original protected URL that the user was trying to access */
 	const char *original_url = apr_pstrdup(r->pool,
@@ -1291,7 +1571,8 @@ static int oidc_handle_post_authorization_response(request_rec *r, oidc_cfg *c,
 			|| ((apr_table_elts(params)->nelts == 1)
 					&& (apr_strnatcmp(apr_table_get(params, "response_mode"),
 							"fragment") == 0))) {
-		return oidc_util_html_send_error(r, "mod_auth_openidc",
+		return oidc_util_html_send_error(r, c->error_template,
+				"Invalid Request",
 				"You've hit an OpenID Connect Redirect URI with no parameters, this is an invalid request; you should not open this URL in your browser directly, or have the server administrator use a different OIDCRedirectURI setting.",
 				HTTP_INTERNAL_SERVER_ERROR);
 	}
@@ -1331,7 +1612,7 @@ static int oidc_discovery(request_rec *r, oidc_cfg *cfg) {
 			&auth_openidc_module);
 
 	/* obtain the URL we're currently accessing, to be stored in the state/session */
-	char *current_url = oidc_get_current_url(r, cfg);
+	char *current_url = oidc_get_current_url(r);
 
 	/* generate CSRF token */
 	char *csrf = NULL;
@@ -1364,7 +1645,8 @@ static int oidc_discovery(request_rec *r, oidc_cfg *cfg) {
 	/* get a list of all providers configured in the metadata directory */
 	apr_array_header_t *arr = NULL;
 	if (oidc_metadata_list(r, cfg, &arr) == FALSE)
-		return oidc_util_html_send_error(r, "mod_auth_openidc",
+		return oidc_util_html_send_error(r, cfg->error_template,
+				"Configuration Error",
 				"No configured providers found, contact your administrator",
 				HTTP_UNAUTHORIZED);
 
@@ -1419,7 +1701,7 @@ static int oidc_discovery(request_rec *r, oidc_cfg *cfg) {
 	/* now send the HTML contents to the user agent */
 	return oidc_util_html_send(r, "OpenID Connect Provider Discovery",
 			"<style type=\"text/css\">body {text-align: center}</style>", NULL,
-			s, HTTP_UNAUTHORIZED);
+			s, DONE);
 }
 
 /*
@@ -1448,6 +1730,23 @@ static int oidc_authenticate_user(request_rec *r, oidc_cfg *c,
 	if (oidc_proto_generate_nonce(r, &nonce, OIDC_PROTO_NONCE_LENGTH) == FALSE)
 		return HTTP_INTERNAL_SERVER_ERROR;
 
+	char *code_verifier = NULL;
+	char *code_challenge = NULL;
+
+	if ((oidc_util_spaced_string_contains(r->pool, provider->response_type,
+			"code") == TRUE) && (provider->pkce_method != NULL)) {
+
+		/* generate the code verifier value that correlates authorization requests and code exchange requests */
+		if (oidc_proto_generate_code_verifier(r, &code_verifier,
+				OIDC_PROTO_CODE_VERIFIER_LENGTH) == FALSE)
+			return HTTP_INTERNAL_SERVER_ERROR;
+
+		/* generate the PKCE code challenge */
+		if (oidc_proto_generate_code_challenge(r, code_verifier,
+				&code_challenge, provider->pkce_method) == FALSE)
+			return HTTP_INTERNAL_SERVER_ERROR;
+	}
+
 	char *method = "get";
 	// TODO: restore method from discovery too or generate state before doing discover (and losing startSSO effect)
 	/*
@@ -1474,6 +1773,9 @@ static int oidc_authenticate_user(request_rec *r, oidc_cfg *c,
 				json_string(provider->response_mode));
 	if (prompt)
 		json_object_set_new(proto_state, "prompt", json_string(prompt));
+	if (code_verifier)
+		json_object_set_new(proto_state, "code_verifier",
+				json_string(code_verifier));
 
 	/* get a hash value that fingerprints the browser concatenated with the random input */
 	char *state = oidc_get_browser_state_hash(r, nonce);
@@ -1509,11 +1811,10 @@ static int oidc_authenticate_user(request_rec *r, oidc_cfg *c,
 			}
 		}
 	} else {
-		char *p = strstr(o_uri.hostname, c->cookie_domain);
-		if ((p == NULL) || (apr_strnatcmp(c->cookie_domain, p) != 0)) {
+		if (!oidc_util_cookie_domain_valid(r_uri.hostname, c->cookie_domain)) {
 			oidc_error(r,
-					"the domain (%s) configured in OIDCCookieDomain does not match the URL hostname (%s) of the URL being accessed (%s): setting \"state\" and \"session\" cookies will not work!!",
-					c->cookie_domain, o_uri.hostname, original_url);
+				"the domain (%s) configured in OIDCCookieDomain does not match the URL hostname (%s) of the URL being accessed (%s): setting \"state\" and \"session\" cookies will not work!!",
+				c->cookie_domain, o_uri.hostname, original_url);
 			return HTTP_INTERNAL_SERVER_ERROR;
 		}
 	}
@@ -1521,7 +1822,7 @@ static int oidc_authenticate_user(request_rec *r, oidc_cfg *c,
 	/* send off to the OpenID Connect Provider */
 	// TODO: maybe show intermediate/progress screen "redirecting to"
 	return oidc_proto_authorization_request(r, provider, login_hint,
-			c->redirect_uri, state, proto_state, id_token_hint,
+			c->redirect_uri, state, proto_state, id_token_hint, code_challenge,
 			auth_request_params);
 }
 
@@ -1541,9 +1842,17 @@ static apr_byte_t oidc_is_discovery_response(request_rec *r, oidc_cfg *cfg) {
  */
 static int oidc_target_link_uri_matches_configuration(request_rec *r,
 		oidc_cfg *cfg, const char *target_link_uri) {
+
 	apr_uri_t o_uri;
-	apr_uri_t r_uri;
 	apr_uri_parse(r->pool, target_link_uri, &o_uri);
+	if (o_uri.hostname == NULL) {
+		oidc_error(r,
+				"could not parse the \"target_link_uri\" (%s) in to a valid URL: aborting.",
+				target_link_uri);
+		return FALSE;
+	}
+
+	apr_uri_t r_uri;
 	apr_uri_parse(r->pool, cfg->redirect_uri, &r_uri);
 
 	if (cfg->cookie_domain == NULL) {
@@ -1572,7 +1881,9 @@ static int oidc_target_link_uri_matches_configuration(request_rec *r,
 	oidc_dir_cfg *dir_cfg = ap_get_module_config(r->per_dir_config,
 			&auth_openidc_module);
 	if (dir_cfg->cookie_path != NULL) {
-		char *p = strstr(o_uri.path, dir_cfg->cookie_path);
+		char *p =
+				(o_uri.path != NULL) ?
+						strstr(o_uri.path, dir_cfg->cookie_path) : NULL;
 		if ((p == NULL) || (p != o_uri.path)) {
 			oidc_error(r,
 					"the path (%s) configured in OIDCCookiePath does not match the URL path (%s) of the \"target_link_uri\" (%s): aborting to prevent an open redirect.",
@@ -1632,14 +1943,16 @@ static int oidc_handle_discovery_response(request_rec *r, oidc_cfg *c) {
 			issuer, target_link_uri, login_hint);
 
 	if (issuer == NULL) {
-		return oidc_util_html_send_error(r, "mod_auth_openidc",
+		return oidc_util_html_send_error(r, c->error_template,
+				"Invalid Request",
 				"Wherever you came from, it sent you here with the wrong parameters...",
 				HTTP_INTERNAL_SERVER_ERROR);
 	}
 
 	if (target_link_uri == NULL) {
 		if (c->default_sso_url == NULL) {
-			return oidc_util_html_send_error(r, "mod_auth_openidc",
+			return oidc_util_html_send_error(r, c->error_template,
+					"Invalid Request",
 					"SSO to this module without specifying a \"target_link_uri\" parameter is not possible because OIDCDefaultURL is not set.",
 					HTTP_INTERNAL_SERVER_ERROR);
 		}
@@ -1649,7 +1962,8 @@ static int oidc_handle_discovery_response(request_rec *r, oidc_cfg *c) {
 	/* do open redirect prevention */
 	if (oidc_target_link_uri_matches_configuration(r, c,
 			target_link_uri) == FALSE) {
-		return oidc_util_html_send_error(r, "mod_auth_openidc",
+		return oidc_util_html_send_error(r, c->error_template,
+				"Invalid Request",
 				"\"target_link_uri\" parameter does not match configuration settings, aborting to prevent an open redirect.",
 				HTTP_UNAUTHORIZED);
 	}
@@ -1667,8 +1981,9 @@ static int oidc_handle_discovery_response(request_rec *r, oidc_cfg *c) {
 		if (oidc_proto_account_based_discovery(r, c, issuer, &issuer) == FALSE) {
 
 			/* something did not work out, show a user facing error */
-			return oidc_util_html_send_error(r, "mod_auth_openidc",
-					"could not resolve the provided account name to an OpenID Connect provider; check your syntax",
+			return oidc_util_html_send_error(r, c->error_template,
+					"Invalid Request",
+					"Could not resolve the provided account name to an OpenID Connect provider; check your syntax.",
 					HTTP_NOT_FOUND);
 		}
 
@@ -1691,7 +2006,7 @@ static int oidc_handle_discovery_response(request_rec *r, oidc_cfg *c) {
 	}
 
 	/* something went wrong */
-	return oidc_util_html_send_error(r, "mod_auth_openidc",
+	return oidc_util_html_send_error(r, c->error_template, "Invalid Request",
 			"Could not find valid provider metadata for the selected OpenID Connect provider; contact the administrator",
 			HTTP_NOT_FOUND);
 }
@@ -1789,7 +2104,8 @@ static int oidc_handle_logout(request_rec *r, oidc_cfg *c, session_rec *session)
 			const char *error_description = apr_psprintf(r->pool,
 					"Logout URL malformed: %s", url);
 			oidc_error(r, "%s", error_description);
-			return oidc_util_html_send_error(r, url, error_description,
+			return oidc_util_html_send_error(r, c->error_template,
+					"Malformed URL", error_description,
 					HTTP_INTERNAL_SERVER_ERROR);
 
 		}
@@ -1801,7 +2117,8 @@ static int oidc_handle_logout(request_rec *r, oidc_cfg *c, session_rec *session)
 							"logout value \"%s\" does not match the hostname of the current request \"%s\"",
 							apr_uri_unparse(r->pool, &uri, 0), r->hostname);
 			oidc_error(r, "%s", error_description);
-			return oidc_util_html_send_error(r, url, error_description,
+			return oidc_util_html_send_error(r, c->error_template,
+					"Invalid Request", error_description,
 					HTTP_INTERNAL_SERVER_ERROR);
 		}
 
@@ -1812,7 +2129,8 @@ static int oidc_handle_logout(request_rec *r, oidc_cfg *c, session_rec *session)
 							"logout value \"%s\" contains illegal \"\n\" or \"\r\" character(s)",
 							url);
 			oidc_error(r, "%s", error_description);
-			return oidc_util_html_send_error(r, url, error_description,
+			return oidc_util_html_send_error(r, c->error_template,
+					"Invalid Request", error_description,
 					HTTP_INTERNAL_SERVER_ERROR);
 		}
 	}
@@ -1887,14 +2205,7 @@ int oidc_handle_jwks(request_rec *r, oidc_cfg *c) {
 
 static int oidc_handle_session_management_iframe_op(request_rec *r, oidc_cfg *c,
 		session_rec *session, const char *check_session_iframe) {
-
 	oidc_debug(r, "enter");
-
-	if (check_session_iframe == NULL) {
-		oidc_debug(r, "no check_session_iframe configured for current OP");
-		return DONE;
-	}
-
 	apr_table_add(r->headers_out, "Location", check_session_iframe);
 	return HTTP_MOVED_TEMPORARILY;
 }
@@ -1980,8 +2291,7 @@ static int oidc_handle_session_management_iframe_rp(request_rec *r, oidc_cfg *c,
 static int oidc_handle_session_management(request_rec *r, oidc_cfg *c,
 		session_rec *session) {
 	char *cmd = NULL;
-	const char *issuer = NULL, *id_token_hint = NULL, *client_id = NULL,
-			*check_session_iframe = NULL;
+	const char *id_token_hint = NULL, *client_id = NULL, *check_session_iframe = NULL;
 	oidc_provider_t *provider = NULL;
 
 	/* get the command passed to the session management handler */
@@ -2006,7 +2316,7 @@ static int oidc_handle_session_management(request_rec *r, oidc_cfg *c,
 			return oidc_handle_session_management_iframe_op(r, c, session,
 					check_session_iframe);
 		}
-		return DONE;
+		return HTTP_NOT_FOUND;
 	}
 
 	/* see if this is a request for the RP iframe */
@@ -2018,15 +2328,14 @@ static int oidc_handle_session_management(request_rec *r, oidc_cfg *c,
 			return oidc_handle_session_management_iframe_rp(r, c, session,
 					client_id, check_session_iframe);
 		}
-		return DONE;
+		oidc_debug(r, "iframe_rp command issued but no client (%s) and/or no check_session_iframe (%s) set", client_id, check_session_iframe);
+		return HTTP_NOT_FOUND;
 	}
 
 	/* see if this is a request check the login state with the OP */
 	if (apr_strnatcmp("check", cmd) == 0) {
 		oidc_session_get(r, session, OIDC_IDTOKEN_SESSION_KEY, &id_token_hint);
-		oidc_session_get(r, session, OIDC_ISSUER_SESSION_KEY, &issuer);
-		if (issuer != NULL)
-			provider = oidc_get_provider_for_issuer(r, c, issuer, FALSE);
+		oidc_get_provider_from_session(r, c, session, &provider);
 		if ((id_token_hint != NULL) && (provider != NULL)) {
 			return oidc_authenticate_user(r, c, provider,
 					apr_psprintf(r->pool, "%s?session=iframe_rp",
@@ -2088,59 +2397,21 @@ static int oidc_handle_refresh_token_request(request_rec *r, oidc_cfg *c,
 		error_code = "no_access_token_match";
 		goto end;
 	}
-	s_access_token = NULL;
-
-	/* get the refresh token that was stored in the session */
-	const char *refresh_token = NULL;
-	oidc_session_get(r, session, OIDC_REFRESHTOKEN_SESSION_KEY, &refresh_token);
-	if (refresh_token == NULL) {
-		oidc_warn(r,
-				"refresh token request handler called but no refresh_token was found in the session");
-		error_code = "no_refresh_token_exists";
-		goto end;
-	}
 
 	/* get a handle to the provider configuration */
-	const char *issuer = NULL;
 	oidc_provider_t *provider = NULL;
-	oidc_session_get(r, session, OIDC_ISSUER_SESSION_KEY, &issuer);
-	if (issuer == NULL) {
-		oidc_error(r, "session corrupted: no issuer found in session");
-		error_code = "session_corruption";
-		goto end;
-	}
-	provider = oidc_get_provider_for_issuer(r, c, issuer, FALSE);
-	if (provider == NULL) {
-		oidc_error(r, "session corrupted: no provider found for issuer: %s",
-				issuer);
+	if (oidc_get_provider_from_session(r, c, session, &provider) == FALSE) {
 		error_code = "session_corruption";
 		goto end;
 	}
 
-	/* elements returned in the refresh response */
-	char *s_id_token = NULL;
-	int expires_in = -1;
-	char *s_token_type = NULL;
-	char *s_refresh_token = NULL;
-
-	/* refresh the tokens by calling the token endpoint */
-	if (oidc_proto_refresh_request(r, c, provider, refresh_token, &s_id_token,
-			&s_access_token, &s_token_type, &expires_in,
-			&s_refresh_token) == FALSE) {
+	/* execute the actual refresh grant */
+	if (oidc_refresh_access_token(r, c, session, provider, NULL) == FALSE) {
 		oidc_error(r, "access_token could not be refreshed");
 		error_code = "refresh_failed";
 		goto end;
 	}
 
-	/* store the new access_token in the session and discard the old one */
-	oidc_session_set(r, session, OIDC_ACCESSTOKEN_SESSION_KEY, s_access_token);
-	oidc_store_access_token_expiry(r, session, expires_in);
-
-	/* if we have a new refresh token (rolling refresh), store it in the session and overwrite the old one */
-	if (s_refresh_token != NULL)
-		oidc_session_set(r, session, OIDC_REFRESHTOKEN_SESSION_KEY,
-				s_refresh_token);
-
 	/* store the session */
 	oidc_session_save(r, session);
 
@@ -2220,7 +2491,7 @@ int oidc_handle_redirect_uri_request(request_rec *r, oidc_cfg *c,
 	}
 
 	/* something went wrong */
-	return oidc_util_html_send_error(r, "mod_auth_openidc",
+	return oidc_util_html_send_error(r, c->error_template, "Invalid Request",
 			apr_psprintf(r->pool,
 					"The OpenID Connect callback URL received an invalid request: %s",
 					r->args), HTTP_INTERNAL_SERVER_ERROR);
@@ -2291,11 +2562,13 @@ static int oidc_check_userid_openidc(request_rec *r, oidc_cfg *c) {
 		case PASS:
 			return OK;
 		case AUTHENTICATE:
+			/* if this is a Javascript path we won't redirect the user and create a state cookie */
+			if (apr_table_get(r->headers_in, "X-Requested-With") != NULL) return HTTP_UNAUTHORIZED;
 			break;
 	}
 
 	/* else: no session (regardless of whether it is main or sub-request), go and authenticate the user */
-	return oidc_authenticate_user(r, c, NULL, oidc_get_current_url(r, c), NULL,
+	return oidc_authenticate_user(r, c, NULL, oidc_get_current_url(r), NULL,
 			NULL, NULL, NULL);
 }
 
@@ -2364,6 +2637,10 @@ authz_status oidc_authz_checker(request_rec *r, const char *require_args, const
 	json_t *claims = NULL, *id_token = NULL;
 	oidc_authz_get_claims_and_idtoken(r, &claims, &id_token);
 
+	/* merge id_token claims (e.g. "iss") in to claims json object */
+	if (claims)
+		oidc_util_json_merge(id_token, claims);
+
 	/* dispatch to the >=2.4 specific authz routine */
 	authz_status rc = oidc_authz_worker24(r, claims ? claims : id_token, require_args);
 
